/* This file was generated by syncqt. */
#ifndef QT_QTEGLFSKMSGBMSUPPORT_VERSION_H
#define QT_QTEGLFSKMSGBMSUPPORT_VERSION_H

#define QTEGLFSKMSGBMSUPPORT_VERSION_STR "6.10.2"

#define QTEGLFSKMSGBMSUPPORT_VERSION 0x0601002

#endif // QT_QTEGLFSKMSGBMSUPPORT_VERSION_H
