// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
// Qt-Security score:significant reason:default

#ifndef QTWIDGETSGLOBAL_H
#define QTWIDGETSGLOBAL_H

#include <QtGui/qtguiglobal.h>
#include <QtWidgets/qtwidgets-config.h>
#include <QtWidgets/qtwidgetsexports.h>

#endif // QTGUIGLOBAL_H
